#!.venv/bin/python

# Scan filesystem to generate a list of files to back up, based on a
# configuration file.  Pass this list to borg to actually create the
# backup.  Execute a notification script on the remote server to
# report the backup status.

import os
import re
import sys
import stat
import pathlib

import typing

import yaml
import wcmatch.glob # type: ignore
import humanfriendly # type: ignore

class Config:
    root: bytes
    max_file_size: typing.Optional[int]
    one_file_system: bool
    exclude_caches: bool
    exclude: list[bytes]
    force_include: list[bytes]
    notify_email: typing.Optional[str]

    def __init__(self, configfile: str):
        # Read config
        with open(configfile, 'r') as f:
            config = yaml.safe_load(f)
            self.root = config['root'].encode()
            self.one_file_system = config.get('one-file-system', False)
            self.exclude_caches = config.get('exclude-caches', False)

            if 'max-file-size' in config:
                self.max_file_size = humanfriendly.parse_size(
                    config['max-file-size'])
            else:
                self.max_file_size = None

            def process_match_list(config_name):
                raw = config.get(config_name, '').encode().split(b'\n')
                pats = []
                # Prepend '**/' to any relative patterns
                for x in raw:
                    if not len(x):
                        continue
                    if x[0] == b'/':
                        pats.append(x)
                    else:
                        pats.append(b'**/' + x)
                return pats

            self.exclude = process_match_list('exclude')
            self.force_include = process_match_list('force_include')

            self.notify_email = config.get('notify-email', None)

        # Compile patterns
        flags = (wcmatch.glob.GLOBSTAR |
                 wcmatch.glob.DOTGLOB |
                 wcmatch.glob.NODOTDIR |
                 wcmatch.glob.EXTGLOB |
                 wcmatch.glob.BRACE)

        # Path matches if it matches at least one regex in "a" and no
        # regex in "b"
        (a, b) = wcmatch.glob.translate(self.exclude, flags=flags)
        self.exclude_re = ([ re.compile(x) for x in a ],
                           [ re.compile(x) for x in b ])

        (a, b) = wcmatch.glob.translate(self.force_include, flags=flags)
        self.force_include_re = ([ re.compile(x) for x in a ],
                                 [ re.compile(x) for x in b ])

    def match_re(self, re: tuple[list[typing.Pattern],
                                 list[typing.Pattern]],
                 path: bytes, is_dir: bool):
        # If it's a directory, try matching against a trailing slash
        # first.
        if is_dir and self.match_re(re, path + b'/', False):
            return True
        # Path matches if it matches at least one regex in
        # re[0] and no regex in re[1].
        for a in re[0]:
            if a.match(path):
                for b in re[1]:
                    if b.match(path):
                        return False
                return True
        return False

class Backup:
    def __init__(self, config: Config, dry_run: bool):
        self.config = config
        self.dry_run = dry_run

        # All logged messages, with severity
        self.logs: list[tuple[str, str]] = []

    def out(self, path: bytes):
        self.outfile.write(path + (b'\n' if self.dry_run else b'\0'))

    def log(self, letter: str, msg: str):
        colors = { 'E': 31, 'W': 33, 'I': 36 };
        if letter in colors:
            c = colors[letter]
        else:
            c = 0
        sys.stderr.write(f"\033[1;{c}m{letter}:\033[22m {msg}\033[0m\n")
        self.logs.append((letter, msg))

    def run(self, outfile: typing.BinaryIO):
        self.outfile = outfile
        # Base should not end with a slash, but full path should
        if self.config.root.endswith(b'/'):
            base = self.config.root[:-1]
            path = self.config.root
        else:
            base = self.config.root
            path = self.config.root + b'/'
        self.scan(base, path)

    def scan(self, base: bytes, path: bytes,
             parent_st: os.stat_result=None):
        """If the given path should be backed up, print it.  If it's
           a directory and its contents should be included, recurse.
        """

        if base.endswith(b'/'):
            raise Exception("base must not end with /")
        relpath = path[len(base):]
        if not relpath.startswith(b'/'):
            raise Exception(f"relative path (from {repr(base)}, {repr(path)})"
                            + f" must start with /")

        # Copy the path in string form, for logging.  Otherwise, we use
        # bytes directly.
        pathstr = path.decode(errors='backslashreplace')

        try:
            st = os.lstat(path)
            is_dir = stat.S_ISDIR(st.st_mode)
            is_reg = stat.S_ISREG(st.st_mode)

            # See if there's a reason to exclude it
            exclude_reason = None

            if self.config.match_re(self.config.exclude_re, relpath, is_dir):
                # Config file says to exclude
                exclude_reason = ('I', f"skipping, excluded by config file")

            elif (self.config.one_file_system
                  and parent_st is not None
                  and is_dir
                  and st.st_dev != parent_st.st_dev):
                # Crosses a mount point
                exclude_reason = ('I', "skipping, on different filesystem")

            elif (self.config.max_file_size
                  and is_reg
                  and (st.st_blocks * 512) > self.config.max_file_size):
                # Too big
                def format_size(n):
                    return humanfriendly.format_size(
                        n, keep_width=True, binary=True)
                a = format_size(st.st_size)
                b = format_size(self.config.max_file_size)
                exclude_reason = ('W', f"file size {a} exceeds limit {b}")

            # If we have a reason to exclude it, stop now unless it's
            # force-included
            force = self.config.match_re(
                self.config.force_include_re, relpath, is_dir)
            if exclude_reason and not force:
                self.log(exclude_reason[0], f"{exclude_reason[1]}: {pathstr}")
                return

            # Print path for Borg
            self.out(path)

            # Process directories
            if is_dir:

                # Skip if it contains CACHEDIR.TAG
                # (mirroring the --exclude-caches borg option)
                if self.config.exclude_caches:
                    try:
                        tag = b'Signature: 8a477f597d28d172789f06886806bc55'
                        with open(path + b'/CACHEDIR.TAG', 'rb') as f:
                            if f.read(len(tag)) == tag:
                                self.log(
                                    'I', f"skipping cache dir: {pathstr}")
                                return
                    except:
                        pass

                # Recurse
                with os.scandir(path) as it:
                    for entry in it:
                        self.scan(base=base, path=entry.path,
                                  parent_st=st)

        except PermissionError as e:
            self.log('E', f"can't read {pathstr}")
            return

def main(argv: list[str]):
    import argparse

    def humansize(string):
        return humanfriendly.parse_size(string)

    parser = argparse.ArgumentParser(
        prog=argv[0],
        description="Back up the local system using borg",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    base = pathlib.Path(__file__).parent
    parser.add_argument('-c', '--config',
                        help="Config file", default=str(base / "config.yaml"))
    parser.add_argument('-b', '--borg',
                        help="Borg command", default=str(base / "borg.sh"))
    parser.add_argument('-n', '--dry-run', action="store_true",
                        help="Just print filenames, don't run borg")

    args = parser.parse_args()
    config = Config(args.config)

    backup = Backup(config, args.dry_run)
    if args.dry_run:
        with open(os.devnull, "wb") as out:
            backup.run(out)
    else:
        backup.run(sys.stdout.buffer)

if __name__ == "__main__":
    import sys
    main(sys.argv)
